@@ -539,11 +539,11 @@ static GLOB_LIST *config_locate(const char *match, const char *type)
 		return g;
 	}
 	/* else */
-	sprintf(match_o, "%s.o", match);
+	snprintf(match_o, sizeof(match_o), "%s.o", match);
 	g = config_lstmod(match_o, type, 0);
 	if (g == NULL || g->pathc == 0) {
 #ifdef CONFIG_USE_ZLIB
-		sprintf(match_o, "%s.o.gz", match);
+		snprintf(match_o, sizeof(match_o), "%s.o.gz", match);
 		g = config_lstmod(match_o, type, 0);
 		if (g == NULL || g->pathc == 0) 
 #endif
@@ -693,6 +693,10 @@ static void config_show()
 			printf("%s %s\n", execs[i].module, execs[i].cmd);
 		}
 	}
+
+	puts("# Miscellaneous file and directory names");
+	for (i = 0; i < gen_file_count; ++i)
+		printf("%s=%s\n", gen_file[i].base, gen_file[i].name);
 }
 
 /*
@@ -768,23 +772,7 @@ static int read_depfile(void)
 	int ret = 0;
 	int line_no;
 	char buf[3000];
-
-	if (depfile == NULL) {
-		/*
-		 * The default value for the depfile parameter is:
-		 *
-		 * depfile=/lib/modules/`uname -r`/modules.dep
-		 *
-		 * If the config file exist but doesn't have a depfile
-		 * specification, the default is used since modprobe
-		 * can't work without one.
-		 */
-		char depfile_tmp[PATH_MAX];
-
-		sprintf(depfile_tmp, "/lib/modules/%s/modules.dep",
-			uts_info.release); /* CHECKME */
-		depfile = xstrdup(depfile_tmp);
-	} /* else: depfile previously defined in config_read() */
+	const char *depfile = gen_file[GEN_DEPFILE].name;
 
 	if ((fin = fopen(depfile, "r")) == NULL) {
 		error("Can't open dependencies file %s (%s)",
@@ -803,14 +791,14 @@ static int read_depfile(void)
 
 	while (fgets_strip(buf, sizeof(buf) - 1, fin, &line_no) != NULL) {
 		NODE *nod;
-		char line[3000];
+		char line[sizeof(buf)];
 		char *pt;
 		char *modname;
 		int len = strlen(buf);
 
 		pt = line + len-1;
 
-		strcpy(line, buf);
+		strcpy(line, buf);	/* safe, same size */
 		while (len > 0 && isspace(*pt)) {
 			*pt-- = '\0';
 			len --;
@@ -833,34 +821,6 @@ static int read_depfile(void)
 		*pt++ = '\0';
 		nod = new_NODE(stripo(modname), modname, &in_depfile);
 
-#if 0 /* Not yet, this needs a lot more thought... */
-		/* Check for extra module info */
-		while (isspace(*pt))
-			pt++;
-		if (*pt == '[') {
-			for (++pt; *pt && *pt != ']'; ++pt) {
-				if (isspace(*pt))
-					continue;
-
-				if (strncmp(pt, "device=", 7) == 0) {
-					char val[100];
-					char *valp = val;
-					pt += 7;
-					while (*pt && !isspace(*pt) && *pt!=']')
-						*valp++ = *pt++;
-					*valp = '\0';
-					if (any_alias(val) == NULL) {
-					}
-				} else if (strncmp(pt, "type=", 5) == 0) {
-					pt += 4; /* ignore, for now */
-				}
-				if (*pt == '\0' || *pt == ']')
-					break;
-			}
-			if (*pt)
-				++pt;
-		}
-#endif
 		/* Parse the list of modules */
 		while (*pt != '\0') {
 			while (isspace(*pt))
@@ -920,6 +880,27 @@ static int call_rmmod(const char *modname, DESC *desc, int complain)
 	if (i == n_module_stat)
 		return 0; /* It is not even loaded! Easy removal job :-) */
 
+	/* The code below uses system() for commands read from the configuration
+	 * file.  It has to use system() because the commands are arbitrary and
+	 * may require shell expansion.
+	 *
+	 * When running as a normal user, you can specify -C config.file and
+	 * that file can contain any commands.  The commands are executed with
+	 * the normal authority of the user, there is no security exposure.
+	 *
+	 * When invoked out of kmod, modprobe runs with root authority.  If the
+	 * kernel lets user parameters through unchanged then the user could
+	 * pass explot strings to modprobe.  The string could start with '-C' to
+	 * specify a user config file, modprobe would then issue user commands
+	 * as root.  To prevent the use of spurious config files, modprobe run
+	 * from kmod runs in safemode and treats the last parameter as a module
+	 * name, even if it starts with '-'.  Therefore the user cannot specify
+	 * any options to modprobe, only a module name.  When invoked from kmod,
+	 * modprobe will only read commands from /etc/modules.conf.  As long as
+	 * that file is secure (and we have to assume that it is) then using
+	 * system() to invoke commands from the config file is safe.
+	 */
+
 	/*
 	 * OK to delete!
 	 */
@@ -927,6 +908,7 @@ static int call_rmmod(const char *modname, DESC *desc, int complain)
 		const char *name = p->item.s;
 		if ((ex = exec_cmd(EXEC_PRE_REMOVE, name)) != NULL) {
 			verbose("# pre-remove %s:\n%s\n", name, ex);
+			/* Trusting commands from config file */
 			if (runit && system(ex) != 0) {
 				if (!quiet && complain)
 					error("pre-remove %s failed", name);
@@ -939,6 +921,7 @@ static int call_rmmod(const char *modname, DESC *desc, int complain)
 		const char *name = p->item.s;
 		if ((ex = exec_cmd(EXEC_REMOVE, kname)) != NULL) {
 			verbose("# remove %s:\n%s\n", name, ex);
+			/* Trusting commands from config file */
 			if (runit && (ret = system(ex)) != 0) {
 				if (!quiet && complain)
 					error("remove %s failed", name);
@@ -959,6 +942,7 @@ static int call_rmmod(const char *modname, DESC *desc, int complain)
 		const char *name = p->item.s;
 		if ((ex = exec_cmd(EXEC_POST_REMOVE, name)) != NULL) {
 			verbose("# post-remove %s:\n%s\n", name, ex);
+			/* Trusting commands from config file */
 			if (runit && system(ex) != 0) {
 				if (!quiet && complain)
 					error("post-remove %s failed", name);
@@ -999,6 +983,7 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 		const char *name = p->item.s;
 		if ((ex = exec_cmd(EXEC_PRE_INSTALL, name)) != NULL) {
 			verbose("# pre-install %s:\n%s\n", name, ex);
+			/* Trusting commands from config file */
 			if (runit && (ret = system(ex)) != 0)
 				error("pre-install %s failed", name);
 		}
@@ -1012,6 +997,7 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 		if ((ex = exec_cmd(EXEC_INSTALL, name)) != NULL) {
 			verbose("# install %s:\n%s\n", name, ex);
 			used_ex = 1;
+			/* Trusting commands from config file */
 			if (runit && (ret = system(ex)) != 0)
 				error("install %s failed", name);
 		}
@@ -1039,11 +1025,19 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 			char *opt = strtok(xopts, " \t");
 
 			while (opt) {
+				if (my_argc >= sizeof(my_argv)/sizeof(my_argv[0])) {
+					error("Too many arguments for insmod");
+					exit(1);
+				}
 				my_argv[my_argc++] = opt;
 				opt = strtok(NULL, " \t");
 			}
 		}
 
+		if (my_argc >= sizeof(my_argv)/sizeof(my_argv[0])) {
+			error("Too many arguments for insmod");
+			exit(1);
+		}
 		my_argv[my_argc++] = desc->nod->str;
 		modopts = my_argc; /* index of first module option */
 		/*
@@ -1073,6 +1067,10 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 					if (j != my_argc)
 						continue;
 					/* else new option */
+					if (my_argc >= sizeof(my_argv)/sizeof(my_argv[0])) {
+						error("Too many arguments for insmod");
+						exit(1);
+					}
 					my_argv[my_argc++] = "-o";
 				} else if ((optend = strchr(opts, '='))) {
 					int l = optend - opts + 1;
@@ -1090,6 +1088,10 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 				} else
 					break;
 
+				if (my_argc >= sizeof(my_argv)/sizeof(my_argv[0])) {
+					error("Too many arguments for insmod");
+					exit(1);
+				}
 				my_argv[my_argc++] = opts;
 			}
 		}
@@ -1125,24 +1127,36 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 					}
 				}
 
-				if (j == my_argc) /* new option */
+				if (j == my_argc) /* new option */ {
+					if (my_argc >= sizeof(my_argv)/sizeof(my_argv[0])) {
+						error("Too many arguments for insmod");
+						exit(1);
+					}
 					my_argv[my_argc++] = options[nopt];
+				}
 			}
 		}
 
-#ifdef COMBINE_modprobe
+		if (my_argc >= sizeof(my_argv)/sizeof(my_argv[0])) {
+			error("Too many arguments for insmod");
+			exit(1);
+		}
+		my_argv[my_argc] = NULL;	/* for execvp */
+
 		if (verbose) {
 			int i;
 			char cmd[2000]; /* hardcoded, but big enough... */
 
 			cmd[0] = '\0';
 			for (i = 0; i < my_argc; ++i) {
-				strcat(cmd, my_argv[i]);
+				xstrcat(cmd, my_argv[i], sizeof(cmd));
 				if (i != my_argc - 1)
-					strcat(cmd, " ");
+					xstrcat(cmd, " ", sizeof(cmd));
 			}
 			verbose("%s\n", cmd);
 		}
+
+#ifdef COMBINE_modprobe
 		if (runit) {
 			int insmod_main(int argc, char **argv);
 			int flag_autoclean_save = flag_autoclean;
@@ -1152,20 +1166,8 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 			flag_autoclean = flag_autoclean_save;
 		}
 #else
-		if (runit || verbose) {
-			int i;
-			char cmd[2000]; /* hardcoded, but big enough... */
-
-			cmd[0] = '\0';
-			for (i = 0; i < my_argc; ++i) {
-				strcat(cmd, my_argv[i]);
-				if (i != my_argc - 1)
-					strcat(cmd, " ");
-			}
-			verbose("%s\n", cmd);
-			if (runit)
-				ret = system(cmd);
-		}
+		if (runit)
+			ret = xsystem(my_argv[0], my_argv);
 #endif /* COMBINE_modprobe */
 		if (!quiet && ret != 0)
 			error("insmod %s failed", desc->nod->str);
@@ -1178,6 +1180,7 @@ static int insmod(DESC *desc, LINK **new_in_kernel, char *options[])
 		const char *name = p->item.s;
 		if ((ex = exec_cmd(EXEC_POST_INSTALL, name)) != NULL) {
 			verbose("# post-install %s:\n%s\n", name, ex);
+			/* Trusting commands from config file */
 			if (runit && (ret = system(ex)) != 0)
 				error("post-install %s failed", name);
 		}
@@ -1356,8 +1359,10 @@ static int modprobe_fromlist(int argc, char *argv[], char *type, int loadall)
 			int alt = 0;
 
 			g = config_locate(name, type);
-			if (!g || g->pathc == 0)
+			if (!g || g->pathc == 0) {
+				ret = -2;	/* unknown module */
 				continue;
+			}
 
 			do {
 				desc = NULL;
@@ -1470,6 +1475,7 @@ int main(int argc, char *argv[])
 	int showconfig = 0;
 	int list = 0;
 	int o;
+	char **env;
 	struct option long_opts[] = {
 		{"type", 1, 0, 't'},
 		{"all", 0, 0, 'a'},
@@ -1490,6 +1496,44 @@ int main(int argc, char *argv[])
 
 	error_file = "modprobe";
 
+	/* Check for safemode before doing anything else, it affects the
+	 * meaning of the parameters.  If the environment contains exactly
+	 * "HOME=/", "TERM=linux", "PATH=/sbin:/usr/sbin:/bin:/usr/bin"
+	 * then assume modprobe was invoked from kmod and default to safemode,
+	 * this should handle existing 2.2 and 2.4 kernels.
+	 *
+	 * If MOD_SAFEMODE=1 is present then use safe mode.
+	 *
+	 * Otherwise assume that we are invoked by a real user, use unsafe
+	 * mode with their authority.
+	 *
+	 * In safe mode the last parameter _must_ be a module name.  No other
+	 * module names or variable=value strings are allowed in safe mode.
+	 */
+	safemode = 1; 
+	for (env = environ; *env; ++env) {
+		if (strcmp(*env, "HOME=/") == 0)
+			continue;
+		if (strcmp(*env, "TERM=linux") == 0)
+			continue;
+		if (strcmp(*env, "PATH=sbin:/usr/sbin:/bin:/usr/bin") == 0)
+			continue;
+		safemode = 0;
+		break;
+	}
+	if (!safemode) {
+		char *p = getenv("MOD_SAFEMODE");
+		if (p) {
+			safemode = strtoul(p, &p, 0);
+			if (*p) {
+				error("Invalid number in MOD_SAFEMODE=%s", getenv("MOD_SAFEMODE"));
+				exit(1);
+			}
+		}
+	}
+	if (safemode)
+		--argc;		/* Do not scan last parameter in getopt */
+
 	while ((o = getopt_long(argc, argv, "t:acdhlnqrvksC:V",
 				&long_opts[0], NULL)) != EOF) {
 		switch (o) {
@@ -1575,6 +1619,22 @@ int main(int argc, char *argv[])
 
 	argc -= optind;
 	argv += optind;
+	if (safemode) {
+		/* Now scan the last parameter, as a module name */
+		if (++argc != 1) {
+			error("Safe mode has more than one parameter");
+			exit(1);
+		}
+		if (**argv == '-') {
+			error("Safe mode parameter starts with '-'");
+			exit(1);
+		}
+		if (strchr(*argv, '=')) {
+			error("Safe mode parameter contains '='");
+			exit(1);
+		}
+	}
+
 	/*
 	 * argv now points to the first non-option argument
 	 * argc is the remaining argument count
