@@ -19,7 +19,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-#ident "$Id: obj_m68k.c 1.3 Sat, 08 Jul 2000 21:46:03 -0700 kaos $"
+#ident "$Id: obj_m68k.c 1.4 Wed, 23 Aug 2000 10:23:26 +1000 kaos $"
 
 #include <stddef.h>
 #include <module.h>
@@ -139,3 +139,9 @@ arch_finalize_section_address(struct obj_file *f, Elf32_Addr base)
     f->sections[i]->header.sh_addr += base;
   return 1;
 }
+
+int
+arch_archdata (struct obj_file *fin, struct obj_section *sec)
+{
+  return 0;
+}
