@@ -19,7 +19,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-#ident "$Id: obj_reloc.c 1.8 Mon, 07 Aug 2000 01:07:12 -0700 kaos $"
+#ident "$Id: obj_reloc.c 1.9 Fri, 08 Sep 2000 16:46:27 +1100 kaos $"
 
 #include <string.h>
 #include <assert.h>
@@ -95,7 +95,7 @@ obj_check_undefineds(struct obj_file *f, int quiet)
 		sym->secidx = SHN_ABS;
 		sym->value = 0;
 	      }
-	    else
+	    else if (sym->r_type) /* assumes R_arch_NONE is 0 on all arch */
 	      {
 		if (!quiet)
 			error("unresolved symbol %s", sym->name);
@@ -332,7 +332,6 @@ obj_relocate (struct obj_file *f, ElfW(Addr) base)
 		}
 
 	      value = obj_symbol_final_value(f, intsym);
-	      intsym->referenced = 1;
 	    }
 
 #if SHT_RELM == SHT_RELA
