@@ -99,18 +99,26 @@ OPT_LIST *aliases;
 static int n_aliases;
 
 char *insmod_opt = NULL;
-char *depfile = NULL;
-char *pcimapfile = NULL;
-char *isapnpmapfile = NULL;
-char *usbmapfile = NULL;
 char *config_file = NULL;	/* Which file was actually used */
 time_t config_mtime;
 int root_check_off = CONFIG_ROOT_CHECK_OFF;	/* Default is modules must be owned by root */
 static char *config_version;	/* Hack for config_add */
 int quick = 0;			/* Option -A */
 
+/* The initialization order must match the gen_file_enum order in config.h */
+struct gen_files gen_file[] = {
+	{"pcimap", NULL, 0},
+	{"isapnpmap", NULL, 0},
+	{"usbmap", NULL, 0},
+	{"dep", NULL, 0},
+};
+
+const int gen_file_count = sizeof(gen_file)/sizeof(gen_file[0]);
+
 int flag_verbose;
 
+unsigned long safemode;
+
 void verbose(const char *ctl,...)
 {
 	if (flag_verbose) {
@@ -127,21 +135,19 @@ void verbose(const char *ctl,...)
  *	Check to see if the existing modules.xxx files need updating,
  *	based on the timestamps of the modules and the config file.
  */
-static time_t dep_mtime;
-static time_t pcimap_mtime;
-static time_t isapnpmap_mtime;
-static time_t usbmap_mtime;
 static int check_update (const char *file, const struct stat *sb)
 {
 	int len = strlen(file);
+	int i;
 
 	if (!S_ISREG(sb->st_mode))
 		return 0;
-	if (sb->st_mtime <= dep_mtime &&
-	    sb->st_mtime <= pcimap_mtime &&
-	    sb->st_mtime <= isapnpmap_mtime &&
-	    sb->st_mtime <= usbmap_mtime)
-		return 0;
+	for (i = 0; i < gen_file_count; ++i) {
+		if (sb->st_mtime > gen_file[i].mtime)
+			break;
+	}
+	if (i == gen_file_count)
+		return 0;	/* All generated files are up to date */
 
 	if (len > 2 && !strcmp(file + len - 2, ".o"))
 		return 1;
@@ -156,8 +162,9 @@ static int check_update (const char *file, const struct stat *sb)
 
 static int need_update (const char *force_ver, const char *base_dir)
 {
-	struct stat statdep, statpcimap, statisapnpmap, statusbmap, tmp;
+	struct stat tmp;
 	char dep[PATH_MAX];
+	int i;
 	uname (&uts_info);
 	if (!force_ver)
 		force_ver = uts_info.release;
@@ -166,29 +173,22 @@ static int need_update (const char *force_ver, const char *base_dir)
 		/* That's just silly. */
 		return 1;
 
-	if (stat (depfile, &statdep) ||
-	    stat (pcimapfile, &statpcimap) ||
-	    stat (isapnpmapfile, &statisapnpmap) ||
-	    stat (usbmapfile, &statusbmap))
-		/* No dependency file yet, so we need to build it. */
-		return 1;
+	for (i = 0; i < gen_file_count; ++i) {
+		if (stat(gen_file[i].name, &tmp))
+			return 1;	/* No dependency file yet, so we need to build it. */
+		gen_file[i].mtime = tmp.st_mtime;
+	}
 
 	if (stat ("/etc/modules.conf", &tmp) &&
 	    stat ("/etc/conf.modules", &tmp))
 		return 1;
 
-	if (tmp.st_mtime > statdep.st_mtime ||
-	    tmp.st_mtime > statpcimap.st_mtime ||
-	    tmp.st_mtime > statisapnpmap.st_mtime ||
-	    tmp.st_mtime > statusbmap.st_mtime)
-		/* Config file is newer. */
-		return 1;
+	for (i = 0; i < gen_file_count; ++i) {
+		if (tmp.st_mtime > gen_file[i].mtime)
+			return 1;	/* Config file is newer. */
+	}
 
-	dep_mtime = statdep.st_mtime;
-	pcimap_mtime = statpcimap.st_mtime;
-	isapnpmap_mtime = statisapnpmap.st_mtime;
-	usbmap_mtime = statusbmap.st_mtime;
-	sprintf (dep, "%s/lib/modules/%s", base_dir, force_ver);
+	snprintf (dep, sizeof(dep), "%s/lib/modules/%s", base_dir, force_ver);
 	return xftw (dep, check_update);
 }
 
@@ -245,7 +245,7 @@ char *fgets_strip(char *buf, int sizebuf, FILE * fin, int *lineno)
 			while (isspace(*pt))
 				pt++;
 			if (pt > buf + 1) {
-				strcpy(buf + 1, pt);
+				strcpy(buf + 1, pt);	/* safe, backward copy */
 				buf[0] = ' ';
 				end -= (int) (pt - buf) - 1;
 			} else if (pt == buf + 1) {
@@ -352,7 +352,7 @@ static void decode_list(int *n, OPT_LIST **list, char *arg, int adding,
 	} else if (!autoclean)
 		(*list)[where].autoclean = 0;
 	pg = (GLOB_LIST *)xmalloc(sizeof(GLOB_LIST));
-	meta_expand(arg2, pg, NULL, version);
+	meta_expand(arg2, pg, NULL, version, ME_ALL);
 	(*list)[where].opts = addlist(prevlist, pg);
 }
 
@@ -389,7 +389,7 @@ static int build_list(char **in, OPT_LIST **out, char *version, int opts)
 		    (*out)[i].autoclean = 0;
 		}
 		pg = (GLOB_LIST *)xmalloc(sizeof(GLOB_LIST));
-		meta_expand(pt, pg, NULL, version);
+		meta_expand(pt, pg, NULL, version, ME_ALL);
 		(*out)[i].name = xstrdup(pn);
 		(*out)[i].opts = pg;
 		free(p);
@@ -399,6 +399,73 @@ static int build_list(char **in, OPT_LIST **out, char *version, int opts)
 	return i;
 }
 
+/* Environment variables can override defaults, testing only */
+static void gen_file_env(struct gen_files *gf)
+{
+        char *e = xmalloc(strlen(gf->base)+5), *p1 = gf->base, *p2 = e;
+	if (safemode)
+		return;
+	while ((*p2++ = toupper(*p1++))) ;
+	strcpy(p2-1, "PATH");	/* safe, xmalloc */
+	if ((p2 = getenv(e)) != NULL) {
+		free(gf->name);
+		gf->name = xstrdup(p2);
+	}
+	free(e);
+}
+
+/* Read a config option for a generated filename */
+static int gen_file_conf(struct gen_files *gf, int assgn, const char *parm, const char *arg)
+{
+
+	int l = strlen(gf->base);
+	if (assgn &&
+	    strncmp(parm, gf->base, l) == 0 &&
+	    strcmp(parm+l, "file") == 0 &&
+	    !gf->name) {
+		gf->name = xstrdup(arg);
+		return(0);
+	}
+	return(1);
+}
+
+/* Check we have a name for a generated file */
+static int gen_file_check(struct gen_files *gf, GLOB_LIST *g,
+			  char *base_dir, char *version)
+{
+	char tmp[PATH_MAX];
+	int ret = 0;
+	if (!gf->name) {
+		/*
+		 * Specification: config file / no xxxfile parameter
+		 * The default value for generated filename xxx is:
+		 *
+		 * xxxfile=/lib/modules/`uname -r`/modules.xxx
+		 *
+		 * If the config file exists but lacks an xxxfile
+		 * specification, the default value is used since
+		 * the system can't work without one.
+		 */
+		snprintf(tmp, sizeof(tmp), "%s/lib/modules/%s/modules.%s",
+			base_dir, version, gf->base);
+		gf->name = xstrdup(tmp);
+	} else { /* xxxfile defined in modules.conf */
+		/*
+		 * If we have a xxxfile definition in the configuration file
+		 * we must resolve any shell meta-chars in its value.
+		 */
+		if (meta_expand(gf->name, g, base_dir, version, ME_ALL))
+			ret = -1;
+		else if (!g->pathv || g->pathv[0] == NULL)
+			ret = -1;
+		else {
+			free(gf->name);
+			gf->name = xstrdup(g->pathv[0]);
+		}
+	}
+	return(ret);
+}
+
 /*
  *	Read the configuration file.
  *	If parameter "all" == 0 then ignore everything except path info
@@ -410,6 +477,7 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 	#define MAX_LEVEL 20
 	FILE *fin;
 	GLOB_LIST g;
+	int i;
 	int assgn;
 	int drop_default_paths = 1;
 	int lineno = 0;
@@ -422,10 +490,6 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 	char *envpath;
 	char *version;
 	char *type;
-	char depfile_tmp[PATH_MAX];
-	char pcimapfile_tmp[PATH_MAX];
-	char isapnpmapfile_tmp[PATH_MAX];
-	char usbmapfile_tmp[PATH_MAX];
 	char **glb;
 	char old_name[] = "/etc/conf.modules";
 	int conf_file_specified = 0;
@@ -579,11 +643,6 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 
 	config_version = xstrdup(version);
 
-	depfile_tmp[0] = '\0';
-	pcimapfile_tmp[0] = '\0';
-	isapnpmapfile_tmp[0] = '\0';
-	usbmapfile_tmp[0] = '\0';
-
 	/* Only read the default entries on the first file */
 	if (depth == 0) {
 		maxpath = 100;
@@ -625,7 +684,7 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 			n_aliases = build_list(aliaslist, &aliases, version, 0);
 
 		/* Order and priority is now: (MODPATH + modules.conf) || (predefs + modules.conf) */
-		if ((envpath = getenv("MODPATH")) != NULL) {
+		if ((envpath = getenv("MODPATH")) != NULL && !safemode) {
 			size_t len;
 			char *p;
 			char *path;
@@ -660,13 +719,13 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 
 			/* The first entry in the path list */
 			modpath[nmodpath].type = xstrdup("boot");
-			sprintf(tmpline, "%s/lib/modules/boot", base_dir);
+			snprintf(tmpline, sizeof(tmpline), "%s/lib/modules/boot", base_dir);
 			modpath[nmodpath].path = xstrdup(tmpline);
 			++nmodpath;
 
 			/* The second entry in the path list, `uname -r` */
 			modpath[nmodpath].type = xstrdup("toplevel");
-			sprintf(tmpline, "%s/lib/modules/%s", base_dir, version);
+			snprintf(tmpline, sizeof(tmpline), "%s/lib/modules/%s", base_dir, version);
 			modpath[nmodpath].path = xstrdup(tmpline);
 			++nmodpath;
 
@@ -676,7 +735,7 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 				if (*k == '.' && ++n == 2)
 					break;
 			}
-			sprintf(tmpline, "%s/lib/modules/%.*s", base_dir,
+			snprintf(tmpline, sizeof(tmpline), "%s/lib/modules/%.*s", base_dir,
 				(/* typecast for Alpha */ int)(k - version), version);
 			modpath[nmodpath].path = xstrdup(tmpline);
 			++nmodpath;
@@ -688,8 +747,8 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 				for (type = tbtype; *type; ++type) {
 					char path[100];
 
-					sprintf(path, "%s%s/%s", base_dir, *pathp, *type);
-					if (meta_expand(path, &g, NULL, version))
+					snprintf(path, sizeof(path), "%s%s/%s", base_dir, *pathp, *type);
+					if (meta_expand(path, &g, NULL, version, ME_ALL))
 						return -1;
 
 					for (glb = g.pathv; glb && *glb; ++glb) {
@@ -705,30 +764,14 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 			}
 		}
 
-		/* DEPPATH is intended for testing only */
-		if ((envpath = getenv("DEPPATH")) != NULL) {
-			strncpy(depfile_tmp, envpath, sizeof(depfile_tmp));
-		}
-
-		/* PCIMAPPATH is intended for testing only */
-		if ((envpath = getenv("PCIMAPPATH")) != NULL) {
-			strncpy(pcimapfile_tmp, envpath, sizeof(pcimapfile_tmp));
-		}
-
-		/* ISAPNPMAPPATH is intended for testing only */
-		if ((envpath = getenv("ISAPNPMAPPATH")) != NULL) {
-			strncpy(isapnpmapfile_tmp, envpath, sizeof(isapnpmapfile_tmp));
-		}
-
-		/* USBMAPPATH is intended for testing only */
-		if ((envpath = getenv("USBMAPPATH")) != NULL) {
-			strncpy(usbmapfile_tmp, envpath, sizeof(usbmapfile_tmp));
-		}
+		/* Environment overrides for testing only, undocumented */
+		for (i = 0; i < gen_file_count; ++i)
+			gen_file_env(gen_file+i);
 
 	}	/* End of depth == 0 */
 
 	if (conf_file ||
-	    ((conf_file = getenv("MODULECONFIG")) != NULL && *conf_file)) {
+	    ((conf_file = getenv("MODULECONFIG")) != NULL && *conf_file && !safemode)) {
 		if (!(fin = fopen(conf_file, "r"))) {
 			error("Can't open %s", conf_file);
 			return -1;
@@ -919,7 +962,7 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 
 			if (strncmp(arg, "-f", 2) == 0) {
 				char *file = next_word(arg);
-				meta_expand(file, &g, NULL, version);
+				meta_expand(file, &g, NULL, version, ME_ALL);
 				if (access(g.pathc ? g.pathv[0] : file, R_OK) == 0)
 					state[level] = !not;
 				else
@@ -943,11 +986,11 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 
 				arg2 = next_word(cmp);
 
-				meta_expand(arg, &g, NULL, version);
+				meta_expand(arg, &g, NULL, version, ME_ALL);
 				if (g.pathc && g.pathv[0])
 					w1 = g.pathv[0];
 
-				meta_expand(arg2, &g2, NULL, version);
+				meta_expand(arg2, &g2, NULL, version, ME_ALL);
 				if (g2.pathc && g2.pathv[0])
 					w2 = g2.pathv[0];
 
@@ -993,7 +1036,7 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 				 *	"" or "0" or "false" => false
 				 *  defined => true
 				 */
-				if (!meta_expand(arg, &g, NULL, version) &&
+				if (!meta_expand(arg, &g, NULL, version, ME_ALL) &&
 				    g.pathc > 0 &&
 				    strcmp(g.pathv[0], "0") != 0 &&
 				    strcmp(g.pathv[0], "false") != 0 &&
@@ -1019,8 +1062,8 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 			char env[PATH_MAX];
 
 			arg2 = next_word(arg);
-			meta_expand(arg2, &g, NULL, version);
-			sprintf(env, "%s=%s", arg, (g.pathc ? g.pathv[0] : ""));
+			meta_expand(arg2, &g, NULL, version, ME_ALL);
+			snprintf(env, sizeof(env), "%s=%s", arg, (g.pathc ? g.pathv[0] : ""));
 			putenv(env);
 			one_err = 0;
 		}
@@ -1029,7 +1072,7 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 		 * include
 		 */
 		if (!assgn && strcmp(parm, "include") == 0) {
-			meta_expand(arg, &g, NULL, version);
+			meta_expand(arg, &g, NULL, version, ME_ALL);
 
 			if (!do_read(all, version, base_dir, g.pathc ? g.pathv[0] : arg, depth+1))
 				one_err = 0;
@@ -1166,78 +1209,6 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 			one_err = 0;
 		}
 
-		/*
-		 * depfile=
-		 */
-		else if (assgn && (strcmp(parm, "depfile") == 0) &&
-			 (depfile_tmp[0] == '\0')) {
-			/*
-			 * The default value for the depfile parameter is:
-			 *
-			 * depfile=/lib/modules/`uname -r`/modules.dep
-			 *
-			 * If the config file exist but doesn't have a depfile
-			 * specification, the default is used since modprobe
-			 * can't work without one.
-			 */
-			strcpy(depfile_tmp, arg);
-			one_err = 0;
-		}
-
-		/*
-		 * pcimapfile=
-		 */
-		else if (assgn && (strcmp(parm, "pcimapfile") == 0) &&
-			 (pcimapfile_tmp[0] == '\0')) {
-			/*
-			 * The default value for the pcimapfile parameter is:
-			 *
-			 * pcimapfile=/lib/modules/`uname -r`/modules.pcimap
-			 *
-			 * If the config file exist but doesn't have a pcimapfile
-			 * specification, the default is used since modprobe
-			 * can't work without one.
-			 */
-			strcpy(pcimapfile_tmp, arg);
-			one_err = 0;
-		}
-
-		/*
-		 * isapnpmapfile=
-		 */
-		else if (assgn && (strcmp(parm, "isapnpmapfile") == 0) &&
-			 (isapnpmapfile_tmp[0] == '\0')) {
-			/*
-			 * The default value for the isapnpmapfile parameter is:
-			 *
-			 * isapnpmapfile=/lib/modules/`uname -r`/modules.isapnpmap
-			 *
-			 * If the config file exist but doesn't have a isapnpmapfile
-			 * specification, the default is used since modprobe
-			 * can't work without one.
-			 */
-			strcpy(isapnpmapfile_tmp, arg);
-			one_err = 0;
-		}
-
-		/*
-		 * usbmapfile=
-		 */
-		else if (assgn && (strcmp(parm, "usbmapfile") == 0) &&
-			 (usbmapfile_tmp[0] == '\0')) {
-			/*
-			 * The default value for the usbmapfile parameter is:
-			 *
-			 * usbmapfile=/lib/modules/`uname -r`/modules.usbmap
-			 *
-			 * If the config file exist but doesn't have a usbmapfile
-			 * specification, the default is used since modprobe
-			 * can't work without one.
-			 */
-			strcpy(usbmapfile_tmp, arg);
-			one_err = 0;
-		}
-
 		/*
 		 * keep
 		 */
@@ -1319,7 +1290,7 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 			/*
 			 * Handle the actual path description
 			 */
-			if (meta_expand(arg, &g, base_dir, version))
+			if (meta_expand(arg, &g, base_dir, version, ME_ALL))
 				return -1;
 			for (glb = g.pathv; glb && *glb; ++glb) {
 				modpath[nmodpath].type = xstrdup(type);
@@ -1333,6 +1304,10 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 			one_err = 0;
 		}
 
+		/* Names for generated files in config file */
+		for (i = 0; one_err && i < gen_file_count; ++i)
+			one_err = gen_file_conf(gen_file+i, assgn, parm, arg);
+
 		/*
 		 * any errors so far?
 		 */
@@ -1352,129 +1327,13 @@ static int do_read(int all, char *force_ver, char *base_dir, char *conf_file, in
 		ret = -1;
 	}
 
-	if (ret == -1)
+	if (ret)
 		return ret;
 	/* else */
 
-	/*
-	 * Make sure we have a definition of "depfile"
-	 */
-	if (depfile_tmp[0] == '\0') {
-		/*
-		 * Specification: config file / no depfile parameter
-		 * The default value for the depfile parameter is:
-		 *
-		 * depfile=/lib/modules/`uname -r`/modules.dep
-		 *
-		 * If the config file exists but lack a depfile
-		 * specification, the default value is used since
-		 * the system can't work without one.
-		 */
-		sprintf(depfile_tmp, "%s/lib/modules/%s/modules.dep",
-			base_dir, version);
-		depfile = xstrdup(depfile_tmp);
-	} else { /* depfile defined in modules.conf */
-		/*
-		 * If we have a depfile definition in the configuration file
-		 * we must resolve any shell meta-chars in its value.
-		 */
-		if (meta_expand(depfile_tmp, &g, base_dir, version))
-			ret = -1;
-		else if (!g.pathv || g.pathv[0] == NULL)
-			ret = -1;
-		else
-			depfile = g.pathv[0];
-	}
-
-	/*
-	 * Make sure we have a definition of "pcimapfile"
-	 */
-	if (pcimapfile_tmp[0] == '\0') {
-		/*
-		 * Specification: config file / no pcimapfile parameter
-		 * The default value for the pcimapfile parameter is:
-		 *
-		 * pcimapfile=/lib/modules/`uname -r`/modules.pcimap
-		 *
-		 * If the config file exists but lack a pcimapfile
-		 * specification, the default value is used since
-		 * the system can't work without one.
-		 */
-		sprintf(pcimapfile_tmp, "%s/lib/modules/%s/modules.pcimap",
-			base_dir, version);
-		pcimapfile = xstrdup(pcimapfile_tmp);
-	} else { /* pcimapfile defined in modules.conf */
-		/*
-		 * If we have a pcimapfile definition in the configuration file
-		 * we must resolve any shell meta-chars in its value.
-		 */
-		if (meta_expand(pcimapfile_tmp, &g, base_dir, version))
-			ret = -1;
-		else if (!g.pathv || g.pathv[0] == NULL)
-			ret = -1;
-		else
-			pcimapfile = g.pathv[0];
-	}
-
-	/*
-	 * Make sure we have a definition of "isapnpmapfile"
-	 */
-	if (isapnpmapfile_tmp[0] == '\0') {
-		/*
-		 * Specification: config file / no isapnpmapfile parameter
-		 * The default value for the isapnpmapfile parameter is:
-		 *
-		 * isapnpmapfile=/lib/modules/`uname -r`/modules.isapnpmap
-		 *
-		 * If the config file exists but lack a isapnpmapfile
-		 * specification, the default value is used since
-		 * the system can't work without one.
-		 */
-		sprintf(isapnpmapfile_tmp, "%s/lib/modules/%s/modules.isapnpmap",
-			base_dir, version);
-		isapnpmapfile = xstrdup(isapnpmapfile_tmp);
-	} else { /* isapnpmapfile defined in modules.conf */
-		/*
-		 * If we have a isapnpmapfile definition in the configuration file
-		 * we must resolve any shell meta-chars in its value.
-		 */
-		if (meta_expand(isapnpmapfile_tmp, &g, base_dir, version))
-			ret = -1;
-		else if (!g.pathv || g.pathv[0] == NULL)
-			ret = -1;
-		else
-			isapnpmapfile = g.pathv[0];
-	}
-
-	/*
-	 * Make sure we have a definition of "usbmapfile"
-	 */
-	if (usbmapfile_tmp[0] == '\0') {
-		/*
-		 * Specification: config file / no usbmapfile parameter
-		 * The default value for the usbmapfile parameter is:
-		 *
-		 * usbmapfile=/lib/modules/`uname -r`/modules.usbmap
-		 *
-		 * If the config file exists but lack a usbmapfile
-		 * specification, the default value is used since
-		 * the system can't work without one.
-		 */
-		sprintf(usbmapfile_tmp, "%s/lib/modules/%s/modules.usbmap",
-			base_dir, version);
-		usbmapfile = xstrdup(usbmapfile_tmp);
-	} else { /* usbmapfile defined in modules.conf */
-		/*
-		 * If we have a usbmapfile definition in the configuration file
-		 * we must resolve any shell meta-chars in its value.
-		 */
-		if (meta_expand(usbmapfile_tmp, &g, base_dir, version))
-			ret = -1;
-		else if (!g.pathv || g.pathv[0] == NULL)
-			ret = -1;
-		else
-			usbmapfile = g.pathv[0];
-	}
+	/* Check we have names for generated files */
+	for (i = 0; !ret && i < gen_file_count; ++i)
+		ret = gen_file_check(gen_file+i, &g, base_dir, version);
 
 	return ret;
 }
@@ -1500,12 +1359,12 @@ int config_read(int all, char *force_ver, char *base_dir, char *conf_file)
 
 /****************************************************************************/
 /*
- *	Add conditionally a file name if it exist
+ *	FIXME: Far too much global state.  KAO.
  */
 static int found;
 static int favail;
 static int one_only;
-static int need_meta_expand;
+static int meta_expand_type;
 char **list;
 static const char *filter_by_file;
 static char *filter_by_dir;
@@ -1519,17 +1378,17 @@ static int config_add(const char *file, const struct stat *sb)
 	int npaths = 0;
 	char **paths = NULL;
 
-	if (need_meta_expand) {
+	if (meta_expand_type) {
 		GLOB_LIST g;
 		char **p;
 		char full[PATH_MAX];
 
-		sprintf(full, "%s/%s", file, filter_by_file);
+		snprintf(full, sizeof(full), "%s/%s", file, filter_by_file);
 
 		if (filter_by_dir && !strstr(full, filter_by_dir))
 			return 0;
 
-		if (meta_expand(full, &g, NULL, config_version))
+		if (meta_expand(full, &g, NULL, config_version, meta_expand_type))
 			return 1;
 		for (p = g.pathv; p && *p; ++p) {
 			paths = (char **)xrealloc(paths,
@@ -1621,7 +1480,7 @@ GLOB_LIST *config_lstmod(const char *match, const char *type, int first_only)
 	struct stat sb;
 	int i;
 	int ret = 0;
-	char *path;
+	char *path = NULL;
 	char this[PATH_MAX];
 
 	if (!match)
@@ -1632,41 +1491,44 @@ GLOB_LIST *config_lstmod(const char *match, const char *type, int first_only)
 	filter_by_dir = NULL;
 	if (type) {
 		char tmpdir[PATH_MAX];
-		sprintf(tmpdir, "/%s/", type);
+		snprintf(tmpdir, sizeof(tmpdir), "/%s/", type);
 		filter_by_dir = xstrdup(tmpdir);
 	}
-	need_meta_expand = strpbrk(match, SHELL_META) && strcmp(match, "*");
+	/* In safe mode, the module name is always handled as is, without meta
+	 * expansion.  It might have come from an end user via kmod and must
+	 * not be trusted.  Even in unsafe mode, only apply globbing to the
+	 * module name, not command expansion.  We trust config file input so
+	 * applying command expansion is safe, we do not trust command line input.
+	 * This assumes that the only time the user can specify -C config file 
+	 * is when they run under their own authority.  In particular all
+	 * mechanisms that call modprobe as root on behalf of the user must
+	 * run in safe mode, without letting the user supply a config filename.
+	 */
+	meta_expand_type = 0;
+	if (strpbrk(match, SHELL_META) && strcmp(match, "*") && !safemode)
+		meta_expand_type = ME_GLOB|ME_BUILTIN_COMMAND;
 
 	list = (char **)xmalloc((favail = 100) * sizeof(char *));
 
 	for (i = 0; i < nmodpath; i++) {
+		path = modpath[i].path;
 		/* Special case: insmod: handle single, non-wildcard match */
 		if (first_only && strpbrk(match, SHELL_META) == NULL) {
 			/* Fix for "2.1.121 syntax */
-			sprintf(this, "%s/%s/%s", modpath[i].path,
+			snprintf(this, sizeof(this), "%s/%s/%s", path,
 						  modpath[i].type, match);
 			if (stat(this, &sb) == 0 &&
 			    config_add(this, &sb))
 				break;
 			/* End fix for "2.1.121 syntax */
 
-			sprintf(this, "%s/%s", modpath[i].path, match);
+			snprintf(this, sizeof(this), "%s/%s", path, match);
 			if (stat(this, &sb) == 0 &&
 			    config_add(this, &sb))
 				break;
 		}
 
 		/* Start looking */
-#if 0
-		/* Fix for "2.1.121 syntax */
-		sprintf(this, "%s/%s", modpath[i].path, modpath[i].type);
-		if (stat(this, &sb) == 0 && S_ISDIR(sb->st_mode))
-			path = this;
-		else
-		/* End fix for "2.1.121 syntax */
-#endif
-			path = modpath[i].path;
-
 		if ((ret = xftw(path, config_add))) {
 			break;
 		}
@@ -1695,11 +1557,11 @@ char *search_module_path(const char *base)
 	if (g == NULL || g->pathc == 0) {
 		char base_o[PATH_MAX];
 
-		sprintf(base_o, "%s.o", base);
+		snprintf(base_o, sizeof(base_o), "%s.o", base);
 		g = config_lstmod(base_o, NULL, 1);
 #ifdef CONFIG_USE_ZLIB
 		if (g == NULL || g->pathc == 0) {
-			sprintf(base_o, "%s.o.gz", base);
+			snprintf(base_o, sizeof(base_o), "%s.o.gz", base);
 			g = config_lstmod(base_o, NULL, 1);
 		}
 #endif
